"""
Copyright (C) 2025, Bo Gan <ganboing@gmail.com>

SPDX-License-Identifer: Apache 2.0
"""

import ctypes
import os

from rvtrace.device import Device, SINK_TYPES, SINK_TYPES_MAP
from rvtrace.timestamp import Timestamp
from rvtrace.helper import get_pmem_range, send_file

TR_TF_CONTROL = 0x0
TR_TF_IMPL = 0x4
TR_TF_SINK_BASE = 0x10
TR_TF_SINK_BASE_HIGH = 0x14
TR_TF_SINK_LIMIT = 0x18
TR_TF_SINK_WP = 0x1c
TR_TF_TS_CONTROL = 0x40
TR_TF_TS_LOW = 0x44
TR_TF_TS_UPPER = 0x48

class TfControlV0Bits(ctypes.LittleEndianStructure):
    _fields_ = [
        ("tfActive",     ctypes.c_uint32, 1), # TF Master Enable/Reset
        ("tfEnable",     ctypes.c_uint32, 1), # TF Enable (SW controlled)
        ("reserved1",    ctypes.c_uint32, 1),
        ("tfEmpty",      ctypes.c_uint32, 1), # TF has flushed all traces
        ("reserved2",    ctypes.c_uint32, 10),
        ("tfStopOnWrap", ctypes.c_uint32, 1), # Set tfEnable=0 on wrap
        ("reserved3",    ctypes.c_uint32, 12),
        ("tfSinkError",  ctypes.c_uint32, 1), # Set when TileLink error (write 1 to clear)
        ("tfSink",       ctypes.c_uint32, 4), # Sink type: 7: SBA (RAM)
    ]

class TfControlV0(ctypes.Union):
    _anonymous_ = ("bits",)
    _fields_ = [
        ("bits",  TfControlV0Bits),
        ("value", ctypes.c_uint32),
    ]

class TfImplV0Bits(ctypes.LittleEndianStructure):
    _fields_ = [
        ("tfVersion",       ctypes.c_uint32, 4), # TF version
        ("tfHasSRAMSink",   ctypes.c_uint32, 1), # TF has on-chip SRAM sink
        ("tfHasATBSink",    ctypes.c_uint32, 1), # TF has ATB sink
        ("tfHasPIBSink",    ctypes.c_uint32, 1), # TF has off-chip trace port via PIB
        ("tfHasSBASink",    ctypes.c_uint32, 1), # TF has on-chip bus master through Front Port
        ("tfHasFunnelSink", ctypes.c_uint32, 1), # TF feeds into a Trace Funnel
        ("reserved1",       ctypes.c_uint32, 7),
        ("tfSinkBytes",     ctypes.c_uint32, 2), # TF trace sink width
        ("reserved2",       ctypes.c_uint32, 14),
    ]

class TfImplV0(ctypes.Union):
    _anonymous_ = ("bits",)
    _fields_ = [
        ("bits",  TfImplV0Bits),
        ("value", ctypes.c_uint32),
    ]

class TraceFunnelV0(Device):
    """
    Pre-ratified version of Trace Funnel.
    (Basically Funnel + Sink in 1.0+)
    """

    @property
    def control(self):
        return self.mmio.read32(TR_TF_CONTROL)

    @control.setter
    def control(self, value):
        self.mmio.write32(TR_TF_CONTROL, value)

    @property
    def tscontrol(self):
        return self.mmio.read32(TR_TF_TS_CONTROL)

    @tscontrol.setter
    def tscontrol(self, value):
        self.mmio.write32(TR_TF_TS_CONTROL, value)

    @property
    def ts(self):
        return (self.mmio.read32(TR_TF_TS_UPPER) << 32) + self.mmio.read32(TR_TF_TS_LOW)

    @ts.setter
    def ts(self, value):
        self.mmio.write32(TR_TF_TS_LOW, value & 0xffffffff)
        self.mmio.write32(TR_TF_TS_UPPER, value >> 32)

    @property
    def impl(self):
        return self.mmio.read32(TR_TF_IMPL)

    @property
    def sinkbase(self):
        return (self.mmio.read32(TR_TF_SINK_BASE_HIGH) << 32) + self.mmio.read32(TR_TF_SINK_BASE)

    @sinkbase.setter
    def sinkbase(self, value):
        self.mmio.write32(TR_TF_SINK_BASE, value & 0xffffffff)
        self.mmio.write32(TR_TF_SINK_BASE_HIGH, value >> 32)

    @property
    def sinklimit(self):
        return (self.mmio.read32(TR_TF_SINK_BASE_HIGH) << 32) + self.mmio.read32(TR_TF_SINK_LIMIT)

    @sinklimit.setter
    def sinklimit(self, value):
        sink_base_high = self.mmio.read32(TR_TF_SINK_BASE_HIGH)
        if sink_base_high != (value >> 32):
            raise Exception(f'SinkLimitHigh {value >> 32:0x} must be equal to '\
                            f'SinkBaseHigh {sink_base_high:0x}')
        self.mmio.write32(TR_TF_SINK_LIMIT, value & 0xffffffff)

    @property
    def sinkwp(self):
        return (self.mmio.read32(TR_TF_SINK_BASE_HIGH) << 32) + self.mmio.read32(TR_TF_SINK_WP)

    @sinkwp.setter
    def sinkwp(self, value):
        sink_base_high = self.mmio.read32(TR_TF_SINK_BASE_HIGH)
        if sink_base_high != (value >> 32):
            raise Exception(f'SinkWPHigh {value >> 32:0x} must be equal to '\
                            f'SinkBaseHigh {sink_base_high:0x}')
        self.mmio.write32(TR_TF_SINK_WP, value & 0xffffffff)

    @staticmethod
    def str_control(control):
        c = TfControlV0(value=control)
        return f'active={c.tfActive} '\
               f'enable={c.tfEnable} '\
               f'empty={c.tfEmpty} '\
               f'stoponwrap={c.tfStopOnWrap} '\
               f'sinkerr={c.tfSinkError} '\
               f'sink={SINK_TYPES[c.tfSink]}'

    @staticmethod
    def str_impl(impl):
        c = TfImplV0(value=impl)
        sink_bytes = ['32-bit', '64-bit', '128-bit', '256-bit']
        return f'version={c.tfVersion} '\
               f'sram={c.tfHasSRAMSink} '\
               f'atb={c.tfHasATBSink} '\
               f'pib={c.tfHasPIBSink} '\
               f'sba={c.tfHasSBASink} '\
               f'funnel={c.tfHasFunnelSink} '\
               f'sinkbytes={sink_bytes[c.tfSinkBytes]}'

    def info(self):
        return f'\timpl: {TraceFunnelV0.str_impl(self.impl)}\n' \
               f'\tcontrol={TraceFunnelV0.str_control(self.control)}\n'\
               f'\tsinkbase={self.sinkbase:0x} '\
               f'sinklimit={self.sinklimit:0x} '\
               f'sinkwp={self.sinkwp:0x}\n'\
               f'\ttimestamp: {Timestamp.str_control(self.tscontrol)}'

    def reset(self):
        super().reset()
        Timestamp.reset(self)

    def probe(self):
        self.reset()
        orig_control = self.control
        supported_sinks = []
        for sink in range(0, 16):
            self.control = TfControlV0(tfActive=1, tfSink=sink).value
            if TfControlV0(value=self.control).tfSink == sink:
                supported_sinks.append(sink)
        self.control = orig_control
        return f'sinks={supported_sinks}'\
               f'\n\ttimstamp: {Timestamp.probe(self)}'

    def config(self):
        enabled = self.runtime.get('enabled', False)
        if not enabled:
            self.control = 0
            return
        #self.control = TfControlV0(tfActive=0).value
        stoponwrap = self.runtime.get('stoponwrap', False)
        output = self.runtime.get('output', None)
        if output is not None and output not in SINK_TYPES_MAP:
            raise Exception(f'unknown output type {output}')
        if output in ['sram', 'sba']:
            pmem = self.runtime.get('pmem', None)
            if pmem is None:
                raise Exception(f'Do not know how to config sink {output}')
            pmem_base, pmem_size = get_pmem_range(pmem)
            if pmem_size == 0:
                raise Exception(f'pmem device {pmem} has zero size')
            align = 2 ** (TfImplV0(value=self.impl).tfSinkBytes + 2)
            if pmem_base % align != 0 or pmem_size % align != 0:
                raise Exception(f'pmem device {pmem} is not aligned to {align} bytes')
            pmem_limit = pmem_base + pmem_size - align
            self.sinkbase = pmem_base
            self.sinklimit = pmem_limit
            self.sinkwp = pmem_base
        self.control = TfControlV0(tfActive=1,
                                   tfStopOnWrap=1 if stoponwrap else 0,
                                   tfSink=0 if output is None else SINK_TYPES_MAP[output]).value
        if 'timestamp' in self.runtime:
            Timestamp.config(self, self.runtime['timestamp'])

    def start(self):
        if 'timestamp' in self.runtime:
            Timestamp.start(self)
        control = TfControlV0(value=self.control)
        control.tfEnable = 1
        control.tfSinkError = 1
        self.control = control.value

    def stop(self):
        control = TfControlV0(value=self.control)
        control.tfEnable = 0
        control.tfSinkError = 0
        self.control = control.value
        while TfControlV0(value=self.control).tfEmpty == 0:
            print(f'{self.name}: waiting for trace flush...')
        if 'timestamp' in self.runtime:
            Timestamp.stop(self)

    def dump(self, file, maxsize):
        control = TfControlV0(value=self.control)
        if control.tfEnable != 0:
            raise Exception(f"Funnel {self.name} must be stopped before dumping")
        output = self.runtime.get('output', None)
        if output not in ['sram', 'sba']:
            raise Exception(f'Funnel {self.name} must be configured in sram/sba mode')
        pmem = self.runtime.get('pmem', None)
        if pmem is None:
            raise Exception('No pmem device configured, unable to dump')
        sinkbase = self.sinkbase
        sinklimit = self.sinklimit
        sinkwp = self.sinkwp
        align = 2 ** (TfImplV0(value=self.impl).tfSinkBytes + 2)
        if sinkwp < sinkbase or sinkwp > sinklimit:
            raise Exception(f'sinkwp {sinkwp:0x} out of range [{sinkbase:0x},{sinklimit:0x}]')
        buffersize = sinklimit - sinkbase + align
        pos = (sinkwp & ~1) - sinkbase
        if pos % align != 0:
            raise Exception(f'pos {pos:0x} is not properly aligned to {align} bytes')
        maxsize = min(maxsize, buffersize)
        with open(f'/dev/{pmem}', 'rb') as pmemf:
            if (sinkwp & 1) == 0 or pos >= maxsize:
                size = min(pos, maxsize)
                send_file(file, pmemf, pos - size, size)
                return size
            # wrap around
            send_file(file, pmemf, buffersize - (maxsize - pos), maxsize - pos)
            send_file(file, pmemf, 0, pos)
            return maxsize



"""
Copyright (C) 2025, Bo Gan <ganboing@gmail.com>

SPDX-License-Identifer: Apache 2.0
"""

import sys
import time
import ctypes
from rvtrace.device import RESET_TIMEOUT

class TsControlBits(ctypes.LittleEndianStructure):
    _fields_ = [
        ("tsActive",    ctypes.c_uint32, 1), # TS Master Enable/Reset
        ("tsCount",     ctypes.c_uint32, 1), # TS Running (internal clk only)
        ("tsReset",     ctypes.c_uint32, 1), # Set 1 to reset timestamp (internal clk only)
        ("tsRunInDbg",  ctypes.c_uint32, 1), # TS Continues in Debug Mode (internal clk only)
        ("tsType",      ctypes.c_uint32, 3), # TS Type (RO)
        ("reserved1",   ctypes.c_uint32, 1),
        ("tsPrescale",  ctypes.c_uint32, 2), # Prescale timestamp by 2^(2*tsPrescale)
        ("reserved2",   ctypes.c_uint32, 5),
        ("tsEnable",    ctypes.c_uint32, 1), # TS/TE Enable timestamp in msg
        ("tsBranch",    ctypes.c_uint32, 2), # TS/TE Add timestamp field to branch messages
        ("tsITC",       ctypes.c_uint32, 1), # TS/TE Add timestamp field to Instrumentation messages
        ("tsOwnership", ctypes.c_uint32, 1), # TS/TE Add timestamp field to Ownership messages
        ("reserved3",   ctypes.c_uint32, 4),
        ("tsWidth",     ctypes.c_uint32, 6), # TS timestamp width
        ("reserved4",   ctypes.c_uint32, 2),
    ]

class TsControl(ctypes.Union):
    _anonymous_ = ("bits",)
    _fields_ = [
        ("bits",  TsControlBits),
        ("value", ctypes.c_uint32),
    ]

class Timestamp:
    @staticmethod
    def str_control(control):
        c = TsControl(value=control)
        types = ["None",  "External", "Internal-Bus", "Internal-Core",
                 "Slave", "Reserved", "Reserved",     "Reserved"]
        return f'active={c.tsActive} '\
               f'running={c.tsCount} '\
               f'type={types[c.tsType]} '\
               f'prescale={c.tsPrescale} '\
               f'msg={c.tsEnable} '\
               f'branch={c.tsBranch} '\
               f'itc={c.tsITC} '\
               f'ownership={c.tsOwnership} '\
               f'width={c.tsWidth}'

    @staticmethod
    def reset(device):
        device.tscontrol = 0
        device.tscontrol = 1
        start = time.monotonic_ns()
        retry = 0
        while (device.tscontrol & 1) == 0:
            if time.monotonic_ns() - start > RESET_TIMEOUT:
                raise Exception(f'Timeout resetting timstamp unit of {device.name}')
        print(f'Device {device.name} timestamp unit reset, nread={retry}', file=sys.stderr)

    @staticmethod
    def probe(device):
        orig_control = device.tscontrol
        ts_in_debug = False
        supported_types = []
        supported_branch_types = []
        support_itc = False
        support_ownership = False
        device.tscontrol = TsControl(tsActive=1, tsRunInDbg=1).value
        if TsControl(value=device.tscontrol).tsRunInDbg != 0:
            ts_in_debug = True
        for tstype in range(0, 8):
            device.tscontrol = TsControl(tsActive=1, tsType=tstype).value
            if TsControl(value=device.tscontrol).tsType == tstype:
                supported_types.append(tstype)
        for branch in range(0, 4):
            device.tscontrol = TsControl(tsActive=1, tsBranch=branch).value
            if TsControl(value=device.tscontrol).tsBranch == branch:
                supported_branch_types.append(branch)
        device.tscontrol = TsControl(tsActive=1, tsITC=1).value
        if TsControl(value=device.tscontrol).tsITC != 0:
            support_itc = True
        device.tscontrol = TsControl(tsActive=1, tsOwnership=1).value
        if TsControl(value=device.tscontrol).tsOwnership != 0:
            support_ownership = True
        device.tscontrol = orig_control
        return f'types={supported_types} '\
               f'runindebug={ts_in_debug} '\
               f'branchtypes={supported_branch_types} '\
               f'itc={support_itc} '\
               f'ownership={support_ownership}'

    @staticmethod
    def config(device, runtime, msgen=False):
        types_map = { 'external' : 1, 'bus' : 2, 'core' : 3, 'slave' : 4 }
        tstype = runtime.get('type', None)
        if tstype is None:
            raise Exception('timestamp type must be set')
        if tstype not in types_map:
            raise Exception(f'unknown timestamp type {tstype}')
        prescale = runtime.get('prescale', 0)
        branch = runtime.get('branch', 0)
        itc = runtime.get('itc', False)
        ownership = runtime.get('ownership', False)
        device.tscontrol = TsControl(tsActive=1,
                                     tsType=types_map[tstype],
                                     tsPrescale=prescale,
                                     tsEnable=1 if msgen else 0,
                                     tsBranch=branch,
                                     tsITC=1 if itc else 0,
                                     tsOwnership=1 if ownership else 0).value

    @staticmethod
    def start(device):
        control = TsControl(value=device.tscontrol)
        control.tsCount = 1
        device.tscontrol = control.value

    @staticmethod
    def stop(device):
        control = TsControl(value=device.tscontrol)
        control.tsCount = 0
        device.tscontrol = control.value



"""
Copyright (C) 2025, Bo Gan <ganboing@gmail.com>

SPDX-License-Identifer: Apache 2.0
"""

import sys
import json
from os import path
from argparse import ArgumentParser
from rvtrace.topology import TraceTopology

BUILTIN_CFG = {
    'SiFive HiFive Premier P550' : 'p550.cfg'
}

def find_builtin_config():
    with open('/sys/firmware/devicetree/base/model', 'r', encoding='ascii') as f:
        model_txt = f.read().rstrip('\x00')
    cfg = BUILTIN_CFG.get(model_txt, None)
    if cfg is not None:
        return path.join(path.dirname(path.realpath(__file__)), f'platforms/{cfg}')
    raise Exception(f'failed to find builtin configuration for {model_txt}')

def cli_trace_info(topo):
    topo.info()

def cli_trace_reset(topo):
    topo.reset()

def cli_trace_probe(topo):
    topo.probe()

def cli_trace_config(topo):
    topo.config()

def cli_trace_timestamp(topo):
    topo.timestamp()

def cli_trace_start(topo):
    topo.start()

def cli_trace_stop(topo):
    topo.stop()

def cli_trace_dump(topo, file, sink=None, maxsize=sys.maxsize):
    if file[0] == '-':
        print(topo.dump(sys.stdout, sink=sink, maxsize=maxsize))
        return
    with open(file[0], 'wb') as fp:
        print(topo.dump(fp, sink=sink, maxsize=maxsize))

def main():
    parser = ArgumentParser(prog='rvtrace')
    parser.add_argument('-c', '--config', dest='config')
    subparsers = parser.add_subparsers(help='subcommand help')

    sub_info = subparsers.add_parser('info', help='information on trace topology')
    sub_info.set_defaults(func=cli_trace_info)

    sub_reset = subparsers.add_parser('reset', help='reset trace topology')
    sub_reset.set_defaults(func=cli_trace_reset)

    sub_probe = subparsers.add_parser('probe', help='probe trace topology')
    sub_probe.set_defaults(func=cli_trace_probe)

    sub_config = subparsers.add_parser('config', help='configure trace topology')
    sub_config.set_defaults(func=cli_trace_config)

    sub_timestamp = subparsers.add_parser('timestamp', help='get timestamps of trace topology')
    sub_timestamp.set_defaults(func=cli_trace_timestamp)

    sub_start = subparsers.add_parser('start', help='start trace')
    sub_start.set_defaults(func=cli_trace_start)

    sub_stop = subparsers.add_parser('stop', help='stop trace')
    sub_stop.set_defaults(func=cli_trace_stop)

    sub_dump = subparsers.add_parser('dump', help='dump trace buffer')
    sub_dump.add_argument('-s', '--sink', dest='sink', help='specify sink device')
    sub_dump.add_argument('-m', '--maxsize', dest='maxsize', type=int,
                          default=sys.maxsize, help='Maximum buffer to dump')
    sub_dump.add_argument('file', nargs=1)
    sub_dump.set_defaults(func=cli_trace_dump)

    options = parser.parse_args()
    func = vars(options).pop('func', None)
    if func is None:
        raise Exception("subcommand not specified")

    config = vars(options).pop('config', None)
    if config is None:
        config = find_builtin_config()
    with open(config, 'r') as f:
        config = json.load(f)
    func(TraceTopology(config), **vars(options))


"""
Copyright (C) 2025, Bo Gan <ganboing@gmail.com>

SPDX-License-Identifer: Apache 2.0
"""
import os
import re
import stat
from os import path, sendfile

PMEM_NS_PATH=re.compile(r'/ndbus(\d+)/region(\d+)/namespace(\d+)\.(\d+)$')

def get_pmem_range(pmem_dev):
    ns_dir = path.realpath(f'/sys/block/{pmem_dev}/device', strict=True)
    if PMEM_NS_PATH.search(ns_dir) is None:
        raise Exception(f'unexpected pmem device path {ns_dir}')
    region_dir = path.dirname(ns_dir)
    with open(path.join(region_dir, 'resource'), 'r', encoding='ascii') as f:
        resource_txt = f.read()
    with open(path.join(region_dir, 'size'), 'r', encoding='ascii') as f:
        size_txt = f.read()
    return int(resource_txt, 0), int(size_txt, 0)

def send_file(file_out, file_in, offset, size):
    statbuf = os.fstat(file_out.fileno())
    if stat.S_ISREG(statbuf.st_mode):
        file_out.truncate(statbuf.st_size + size)
    while size > 0:
        xfer = min(size, 2 ** 30) # At most 1G a time
        xfer = sendfile(file_out.fileno(), file_in.fileno(), offset, xfer)
        if xfer == 0:
            raise Exception(f'EOF when sendfile {file_in.name} -> {file_out.name}')
        size = size - xfer
        offset = offset + xfer


"""
Reimplementation of the standard Python mmap module with better control over access width,
to support memory-mapped I/O.

Copyright (C) ARM Ltd. 2019.  All rights reserved.
Copyright (C) 2025, Bo Gan <ganboing@gmail.com>

SPDX-License-Identifer: Apache 2.0
"""

import os
import ctypes
import errno
from importlib.machinery import EXTENSION_SUFFIXES
path = os.path

# Import the Python mmap module to get access to its constants.
import mmap as real_mmap
for x in real_mmap.__dict__:
    if x.startswith("PROT_") or x.startswith("MAP_"):
        globals()[x] = real_mmap.__dict__[x]
assert PROT_READ == real_mmap.PROT_READ
del real_mmap

libc = ctypes.CDLL(None, use_errno=True)
libc_mmap = libc.mmap
libc_mmap.restype = ctypes.c_void_p
libc_mmap.argtypes = [ctypes.c_void_p, ctypes.c_size_t, ctypes.c_int,
                      ctypes.c_int, ctypes.c_int, ctypes.c_ulonglong]
libc_munmap = libc.munmap
libc_munmap.restype = ctypes.c_int
libc_munmap.argtypes = [ctypes.c_void_p, ctypes.c_size_t]

def find_library(name):
    filedir = path.dirname(path.realpath(__file__))
    suffixes = [*EXTENSION_SUFFIXES]
    assert(len(suffixes) > 0)
    while True:
        try:
            extpath = path.join(filedir, f"../{name}{suffixes.pop(0)}")
            lib = ctypes.CDLL(extpath)
        except Exception as e:
            if len(suffixes) == 0:
                raise Exception(f"Cannot load {name}, last error: {e}")
            continue
        return lib

libmmio = find_library('libmmio')
for size in [8, 16, 32, 64]:
    func = getattr(libmmio, f'mmio_read{size}')
    globals()[f'libmmio_read{size}'] = func
    func.restype = getattr(ctypes, f'c_uint{size}')
    func.argtypes = [ctypes.c_void_p]
    func = getattr(libmmio, f'mmio_write{size}')
    globals()[f'libmmio_write{size}'] = func
    func.restype = ctypes.c_void_p
    func.argtypes = [ctypes.c_void_p, getattr(ctypes, f'c_uint{size}')]

class mmap:
    """
    Represent a single block of memory allocated by mmap.
    """
    def __init__(self, fno, size, flags=MAP_SHARED, prot=(PROT_WRITE|PROT_READ), offset=0):
        assert size > 0
        assert (size % os.sysconf("SC_PAGE_SIZE")) == 0
        assert offset >= 0
        self.size = size
        self.addr = libc_mmap(0, size, prot, flags, fno, offset)
        if (self.addr & 0xfff) == 0xfff:
            # Mapping failed. Possible reasons:
            #  - CONFIG_IO_STRICT_DEVMEM and area is forbidden
            if ctypes.get_errno() == errno.EPERM:
                raise PermissionError
            raise EnvironmentError

    def close(self):
        if libc_munmap(self.addr, self.size) != 0:
            raise OSError

    def read8(self, pos):
        return self.read(pos, 1)

    def read16(self, pos):
        return self.read(pos, 2)

    def read32(self, pos):
        return self.read(pos, 4)

    def read64(self, pos):
        return self.read(pos, 8)

    def read(self, start, nbytes):
        assert nbytes in [1,2,4,8], "invalid length for read: %u" % nbytes
        if nbytes == 1:
            return libmmio_read8(self.addr + start)
        elif nbytes == 2:
            return libmmio_read16(self.addr + start)
        elif nbytes == 4:
            return libmmio_read32(self.addr + start)
        elif nbytes == 8:
            return libmmio_read64(self.addr + start)
        else:
            assert False

    def write8(self, pos, value):
        return self.write(pos, 1, value)

    def write16(self, pos, value):
        return self.write(pos, 2, value)

    def write32(self, pos, value):
        return self.write(pos, 4, value)

    def write64(self, pos, value):
        return self.write(pos, 8, value)

    def write(self, start, nbytes, value):
        assert nbytes in [1,2,4,8], "invalid length for write: %u" % nbytes
        if nbytes == 1:
            libmmio_write8(self.addr + start, value)
        elif nbytes == 2:
            libmmio_write16(self.addr + start, value)
        elif nbytes == 4:
            libmmio_write32(self.addr + start, value)
        elif nbytes == 8:
            libmmio_write64(self.addr + start, value)
        else:
            assert False



"""
Copyright (C) 2025, Bo Gan <ganboing@gmail.com>

SPDX-License-Identifer: Apache 2.0
"""

import sys
from rvtrace.encoder import TraceEncoderV0
from rvtrace.funnel import TraceFunnelV0

class TraceEncoderPad:
    def __init__(self, encoder):
        self.encoder = encoder
        self.outputs = []

    @property
    def device(self):
        return self.encoder

class TraceFunnelPad:
    def __init__(self, funnel):
        self.funnel = funnel
        self.outputs = []
        self.inputs = []

    @property
    def device(self):
        return self.funnel

class TraceSinkPad:
    def __init__(self, sink):
        self.sink = sink
        self.inputs = []

    @property
    def device(self):
        return self.sink

DEVICE_TYPES = {
    "riscv,encoder0" : (TraceEncoderV0, TraceEncoderPad),
    "riscv,funnel0" :  (TraceFunnelV0, TraceFunnelPad),
}

class TraceTopology:
    def __init__(self, config):
        self.devices = {}
        self.pads = {}
        self.encoders = []
        self.sinks = []
        for name, cfg in config.items():
            type = cfg.get('type', None)
            if type is None or type not in DEVICE_TYPES:
                raise Exception(f'type must be specified as {DEVICE_TYPES.keys()}')
            reg = cfg.get('reg', None)
            if reg is None:
                raise Exception('reg must be specified')
            device = DEVICE_TYPES[type][0](name=name,
                                           mmio_base=int(reg, 16),
                                           runtime=cfg.get('runtime', {}))
            self.devices[name] = device
            pad = DEVICE_TYPES[type][1](device)
            self.pads[name] = pad
        for name, cfg in config.items():
            pad = self.pads[name]
            outputs = cfg.get('outputs', [])
            for dev in outputs:
                if not hasattr(pad, 'outputs'):
                    raise Exception(f'device {name} should not have outputs')
                if dev not in self.pads:
                    raise Exception(f'output device {dev} not found')
                outpad = self.pads[dev]
                if not hasattr(outpad, 'inputs'):
                    raise Exception(f'device {name} does not have inputs')
                pad.outputs.append(outpad)
                outpad.inputs.append(pad)
        for pad in self.pads.values():
            if not hasattr(pad, 'inputs'):
                self.encoders.append(pad)
            if not hasattr(pad, 'outputs') or not pad.outputs:
                self.sinks.append(pad)

    @staticmethod
    def apply(pads, action):
        if not pads:
            return
        outputs = set()
        for pad in pads:
            action(pad)
            if hasattr(pad, 'outputs'):
                outputs.update(pad.outputs)
        TraceTopology.apply(outputs, action)

    @staticmethod
    def apply_reverse(pads, action):
        if not pads:
            return
        inputs = set()
        for pad in pads:
            action(pad)
            if hasattr(pad, 'inputs'):
                inputs.update(pad.inputs)
        TraceTopology.apply_reverse(inputs, action)

    @staticmethod
    def devprefix(device):
        return f'{device.name} {device.__class__.__name__} @{device.mmio_base:0x}:'

    def info(self, file=sys.stdout):
        def _info(pad):
            device = pad.device
            print(TraceTopology.devprefix(device), file=file)
            print(device.info(), file=file)
            if hasattr(pad, 'inputs'):
                print(f'\tinputs={[input.device.name for input in pad.inputs]}', file=file)
            if hasattr(pad, 'outputs'):
                print(f'\toutputs={[output.device.name for output in pad.outputs]}', file=file)
        TraceTopology.apply(self.encoders, _info)

    def reset(self):
        def _reset(pad):
            device = pad.device
            device.reset()
        TraceTopology.apply(self.encoders, _reset)

    def probe(self, file=sys.stdout):
        def _probe(pad):
            device = pad.device
            print(TraceTopology.devprefix(device), file=file)
            print(f'\t{device.probe()}', file=file)
        TraceTopology.apply(self.encoders, _probe)

    def config(self, file=sys.stdout):
        self.reset()
        def _config(pad):
            device = pad.device
            device.config()
            print(f'{TraceTopology.devprefix(device)} configured', file=file)
        TraceTopology.apply_reverse(self.sinks, _config)

    def timestamp(self, file=sys.stdout):
        def _timestamp(pad):
            device = pad.device
            print(f'{TraceTopology.devprefix(device)} TS={device.ts:016d}', file=file)
        TraceTopology.apply(self.encoders, _timestamp)

    def start(self):
        self.config()
        def _start(pad):
            device = pad.device
            device.start()
        TraceTopology.apply_reverse(self.sinks, _start)

    def stop(self):
        def _stop(pad):
            device = pad.device
            device.stop()
        TraceTopology.apply(self.encoders, _stop)

    def dump(self, fp, sink=None, maxsize=sys.maxsize):
        if sink is None and len(self.sinks) != 1:
            raise Exception('sink must be specified for topology with multiple sinks')
        sink = self.sinks[0] if sink is None else self.pads[sink]
        return sink.device.dump(fp, maxsize)


"""
Copyright (C) 2025, Bo Gan <ganboing@gmail.com>

SPDX-License-Identifer: Apache 2.0
"""

import os
import sys
import mmap
import time
from rvtrace.iommap import mmap, PROT_READ, PROT_WRITE

DEVMEM = "/dev/mem"
TR_COMMON_CONTROL = 0x0
TR_COMMON_IMPL = 0x4
RESET_TIMEOUT = 1000 * 1000 * 1000
SINK_TYPES = ["Default",  "Reserved", "Reserved", "Reserved",
              "SRAM",     "ATB",      "PIB",      "SBA",
              "FUNNEL",   "Reserved", "Reserved", "Reserved",
              "Reserved", "Reserved", "Reserved", "Reserved", ]
SINK_TYPES_MAP = {
    "sram" : 4,
    "atb" : 5,
    "pib" : 6,
    "sba" : 7,
    "funnel" : 8
}

class DevMem(mmap):
    """
    Interface to /dev/mem
    """
    def __init__(self, size, offset=0, writable=False):
        fd = os.open(DEVMEM, os.O_RDWR if writable else os.O_RDONLY)
        prot = PROT_READ | (PROT_WRITE if writable else 0)
        super().__init__(fd, size, prot=prot, offset=offset)
        os.close(fd)

class Device:
    """
    Common for all trace components
    """
    def __init__(self, name, mmio_base, runtime):
        self.name = name
        self.mmio_base = mmio_base
        self.mmio = DevMem(0x1000, offset=mmio_base, writable=True)
        self.tr_impl = self.mmio.read32(TR_COMMON_IMPL)
        self.ver_major = self.tr_impl & 0xf
        self.ver_minor = self.tr_impl & 0xff >> 4
        if self.ver_major > 0:
            self.comp_type = self.tr_impl & 0xfff >> 8
        self.runtime = runtime

    def info(self):
        pass

    def reset(self):
        retry = 0
        self.control = 0
        start = time.monotonic_ns()
        while (self.control & 1) != 0:
            retry = retry + 1
            if time.monotonic_ns() - start > RESET_TIMEOUT:
                raise Exception(f'Timeout resetting {self.name}')
        self.control = 1
        start = time.monotonic_ns()
        while (self.control & 1) == 0:
            retry = retry + 1
            if time.monotonic_ns() - start > RESET_TIMEOUT:
                raise Exception(f'Timeout resetting {self.name}')
        print(f'Device {self.name} reset, control={self.control:0x} nread={retry}', file=sys.stderr)

    def probe(self):
        pass

    def config(self):
        pass

    def start(self):
        pass

    def stop(self):
        pass

    def dump(self, file, maxsize):
        pass



from setuptools import setup, find_packages, Extension

setup(
    name='rvtrace',
    version='0.1.20250114',
    description='User Mode driver for RISC-V Trace',
    url='https://github.com/ganboing/riscv-trace-umd',
    author='Bo Gan',
    author_email='ganboing@gmail.com',
    packages=find_packages(include='rvtrace.*'),
    package_data={'': ['*.cfg', '*.md']},
    entry_points={
        'console_scripts': ['rvtrace=rvtrace.cli:main']
    },
    ext_modules=[
        Extension(name='libmmio', sources=['lib/mmio.c'])
    ]
)



"""
Copyright (C) 2025, Bo Gan <ganboing@gmail.com>

SPDX-License-Identifer: Apache 2.0
"""

import ctypes
from rvtrace.device import Device, SINK_TYPES, SINK_TYPES_MAP
from rvtrace.timestamp import Timestamp

TR_TE_CONTROL = 0x0
TR_TE_IMPL = 0x4
TR_TE_TS_CONTROL = 0x40
TR_TE_TS_LOW = 0x44
TR_TE_TS_UPPER = 0x48
TR_TE_ITC_TRACE_EN = 0x60
TR_TE_ITC_TRIG_EN = 0x64
TR_TE_ITC_BLOCK = 0x80

class TeControlV0Bits(ctypes.LittleEndianStructure):
    _fields_ = [
        ("teActive",      ctypes.c_uint32, 1), # TE Master Enable/Reset
        ("teEnable",      ctypes.c_uint32, 1), # TE Enable (SW controlled)
        ("teTracing",     ctypes.c_uint32, 1), # TE Tracing (SW/HW controlled)
        ("teEmpty",       ctypes.c_uint32, 1), # TE has flushed all traces
        ("teInstMode",    ctypes.c_uint32, 3), # Instruction tracing mode
        ("teITCMode",     ctypes.c_uint32, 2), # ITC trace generation parameters
        ("resreved1",     ctypes.c_uint32, 3),
        ("teStallOvf",    ctypes.c_uint32, 1), # Overflow/stall indicator (SW clearable)
        ("teStallEn",     ctypes.c_uint32, 1), # Force core stall by TE
        ("teStopOnWrap",  ctypes.c_uint32, 1), # Set teEnable=0 on wrap
        ("teInhibitSrc",  ctypes.c_uint32, 1), # Disable SRC field in Nexus msg
        ("teSyncMaxBTM",  ctypes.c_uint32, 4), # SYNC msg per 2^(teSyncMaxBTM+5) BTMs 0xf=OFF
        ("teSyncMaxInst", ctypes.c_uint32, 4), # SYNC msg per 2^(teSyncMaxInst+5) I-CNT 0xf=OFF
        ("teSliceFormat", ctypes.c_uint32, 3), # Set to 1: 6 MDO + 2 MSEO
        ("teSinkError",   ctypes.c_uint32, 1), # Set when sink refuses msgs (write 1 to clear)
        ("teSink",        ctypes.c_uint32, 4), # Sink type: 8: Funnel
    ]

class TeControlV0(ctypes.Union):
    _anonymous_ = ("bits",)
    _fields_ = [
        ("bits",  TeControlV0Bits),
        ("value", ctypes.c_uint32),
    ]

class TeImplV0Bits(ctypes.LittleEndianStructure):
    _fields_ = [
        ("teVersion",       ctypes.c_uint32, 4), # TE version
        ("teHasSRAMSink",   ctypes.c_uint32, 1), # TE has on-chip SRAM sink
        ("teHasATBSink",    ctypes.c_uint32, 1), # TE has ATB sink
        ("teHasPIBSink",    ctypes.c_uint32, 1), # TE has off-chip trace port via PIB
        ("teHasSBASink",    ctypes.c_uint32, 1), # TE has on-chip bus master through Front Port
        ("teHasFunnelSink", ctypes.c_uint32, 1), # TE feeds into a Trace Funnel
        ("reserved1",       ctypes.c_uint32, 7),
        ("teSinkBytes",     ctypes.c_uint32, 2), # TE trace sink width
        ("teCrossingType",  ctypes.c_uint32, 2), # TE clock crossing between ingress and BTM Gen
        ("teSrcID",         ctypes.c_uint32, 4), # TE source ID
        ("teNSrcBits",      ctypes.c_uint32, 3), # TE Nexus SRC field width
        ("reserved2",       ctypes.c_uint32, 1),
        ("teHartID",        ctypes.c_uint32, 4), # TE Hart ID
    ]

class TeImplV0(ctypes.Union):
    _anonymous_ = ("bits",)
    _fields_ = [
        ("bits",  TeImplV0Bits),
        ("value", ctypes.c_uint32),
    ]

class TraceEncoderV0(Device):
    """
    Pre-ratified version of Trace Encoder.
    """
    @property
    def control(self):
        return self.mmio.read32(TR_TE_CONTROL)

    @control.setter
    def control(self, value):
        self.mmio.write32(TR_TE_CONTROL, value)

    @property
    def tscontrol(self):
        return self.mmio.read32(TR_TE_TS_CONTROL)

    @tscontrol.setter
    def tscontrol(self, value):
        self.mmio.write32(TR_TE_TS_CONTROL, value)

    @property
    def ts(self):
        return (self.mmio.read32(TR_TE_TS_UPPER) << 32) + self.mmio.read32(TR_TE_TS_LOW)

    @ts.setter
    def ts(self, value):
        self.mmio.write32(TR_TE_TS_LOW, value & 0xffffffff)
        self.mmio.write32(TR_TE_TS_UPPER, value >> 32)

    @property
    def impl(self):
        return self.mmio.read32(TR_TE_IMPL)

    @staticmethod
    def str_control(control):
        c = TeControlV0(value=control)
        return f'active={c.teActive} '\
               f'enable={c.teEnable} '\
               f'tracing={c.teTracing} '\
               f'empty={c.teEmpty} '\
               f'inst_mode={c.teInstMode} '\
               f'itc_mode={c.teITCMode} '\
               f'stallovf={c.teStallOvf} '\
               f'stallen={c.teStallEn} '\
               f'stoponwrap={c.teStopOnWrap} '\
               f'nosrc={c.teInhibitSrc} '\
               f'syncmaxbtm={c.teSyncMaxBTM} '\
               f'syncmaxinst={c.teSyncMaxInst} '\
               f'nexusformat={c.teSliceFormat} '\
               f'sinkerr={c.teSinkError} '\
               f'sink={SINK_TYPES[c.teSink]}'

    @staticmethod
    def str_impl(impl):
        c = TeImplV0(value=impl)
        sink_bytes = ['32-bit', '64-bit', '128-bit', '256-bit']
        crossing_type = ['Synchronous', 'Rational', 'Asynchronous', 'Unknown']
        return f'version={c.teVersion} '\
               f'sram={c.teHasSRAMSink} '\
               f'atb={c.teHasATBSink} '\
               f'pib={c.teHasPIBSink} '\
               f'sba={c.teHasSBASink} '\
               f'funnel={c.teHasFunnelSink} '\
               f'sinkbytes={sink_bytes[c.teSinkBytes]} '\
               f'crossing_type={crossing_type[c.teCrossingType]} '\
               f'srcid={c.teSrcID} '\
               f'srcbits={c.teNSrcBits} '\
               f'hartid={c.teHartID}'

    def info(self):
        return f'\timpl: {TraceEncoderV0.str_impl(self.impl)}\n' \
               f'\tcontrol: {TraceEncoderV0.str_control(self.control)}\n'\
               f'\ttimestamp: {Timestamp.str_control(self.tscontrol)}'

    def reset(self):
        super().reset()
        Timestamp.reset(self)

    def probe(self):
        self.reset()
        orig_control = self.control
        supported_inst = []
        supported_itc = []
        supported_sync_maxbtm = []
        supported_sync_maxinst = []
        supported_sinks = []
        for inst_mode in range(0, 8):
            self.control = TeControlV0(teActive=1, teInstMode=inst_mode).value
            if TeControlV0(value=self.control).teInstMode == inst_mode:
                supported_inst.append(inst_mode)
        for itc_mode in range(0, 4):
            self.control = TeControlV0(teActive=1, teITCMode=itc_mode).value
            if TeControlV0(value=self.control).teITCMode == itc_mode:
                supported_itc.append(itc_mode)
        for sync_maxbtm in range(0, 16):
            self.control = TeControlV0(teActive=1, teSyncMaxBTM=sync_maxbtm).value
            if TeControlV0(value=self.control).teSyncMaxBTM == sync_maxbtm:
                supported_sync_maxbtm.append(sync_maxbtm)
        for sync_maxinst in range(0, 16):
            self.control = TeControlV0(teActive=1, teSyncMaxInst=sync_maxinst).value
            if TeControlV0(value=self.control).teSyncMaxInst == sync_maxinst:
                supported_sync_maxinst.append(sync_maxinst)
        for sink in range(0, 16):
            self.control = TeControlV0(teActive=1, teSink=sink).value
            if TeControlV0(value=self.control).teSink == sink:
                supported_sinks.append(sink)
        self.control = orig_control
        return f'instmode={supported_inst} '\
               f'itcmode={supported_itc} '\
               f'sync_maxbtm={supported_sync_maxbtm} '\
               f'sync_maxinst={supported_sync_maxinst} '\
               f'sinks={supported_sinks}'\
               f'\n\ttimstamp: {Timestamp.probe(self)}'

    def config(self):
        enabled = self.runtime.get('enabled', False)
        if not enabled:
            self.control = 0
            return
        instmode = self.runtime.get('instmode', 0)
        itcmode = self.runtime.get('itcmode', 0)
        stall = self.runtime.get('stall', False)
        stoponwrap = self.runtime.get('stoponwrap', False)
        syncmaxbtm = self.runtime.get('syncmaxbtm', 15)
        syncmaxinst = self.runtime.get('syncmaxinst', 15)
        output = self.runtime.get('output', None)
        if output is not None and output not in SINK_TYPES_MAP:
            raise Exception(f'unknown output type {output}')
        if output in ['sram', 'sba']:
            raise Exception(f'output type {output} is not supported for encoderv0')
        self.control = TeControlV0(teActive=1,
                                   teInstMode=instmode,
                                   teITCMode=itcmode,
                                   teStallEn=1 if stall else 0,
                                   teStopOnWrap=1 if stoponwrap else 0,
                                   teSyncMaxBTM=syncmaxbtm,
                                   teSyncMaxInst=syncmaxinst,
                                   teSink=0 if output is None else SINK_TYPES_MAP[output]).value
        if 'timestamp' in self.runtime:
            Timestamp.config(self, self.runtime['timestamp'], msgen=True)

    def start(self):
        if 'timestamp' in self.runtime:
            Timestamp.start(self)
        control = TeControlV0(value=self.control)
        control.teEnable = 1
        control.teTracing = 1
        control.teSinkError = 1
        self.control = control.value

    def stop(self):
        control = TeControlV0(value=self.control)
        control.teEnable = 0
        control.teSinkError = 0
        self.control = control.value
        while TeControlV0(value=self.control).teEmpty == 0:
            print(f'{self.name}: waiting for trace flush...')
        if 'timestamp' in self.runtime:
            Timestamp.stop(self)

    def dump(self, *args, **kwargs):
        control = TeControlV0(value=self.control)
        if control.teEnable != 0:
            raise Exception(f"Encoder {self.name} must be stopped before dumping")
        if self.runtime.get('output', None) != 'funnel':
            raise Exception("Only support funnel output of envoderv0")
        raise Exception(f"You should dump through funnel of Encoder {self.name}")
